package sun.nio.ch;

import java.nio.channels.Selector;
import java.util.Iterator;
import java.net.SocketException;
import java.nio.channels.IllegalSelectorException;
import java.nio.channels.spi.AbstractSelectableChannel;
import java.io.IOException;
import java.nio.channels.ClosedSelectorException;
import java.util.Collections;
import java.nio.channels.spi.SelectorProvider;
import java.util.HashSet;
import java.nio.channels.SelectionKey;
import java.util.Set;
import java.nio.channels.spi.AbstractSelector;

public abstract class SelectorImpl extends AbstractSelector
{
    protected Set<SelectionKey> selectedKeys;
    protected HashSet<SelectionKey> keys;
    private Set<SelectionKey> publicKeys;
    private Set<SelectionKey> publicSelectedKeys;
    
    protected SelectorImpl(SelectorProvider provider) {
        super(provider);
        this.keys = new HashSet();
        this.selectedKeys = new HashSet();
        if (Util.atBugLevel("1.4")) {
            this.publicKeys = this.keys;
            this.publicSelectedKeys = this.selectedKeys;
        }
        else {
            this.publicKeys = Collections.unmodifiableSet(this.keys);
            this.publicSelectedKeys = Util.ungrowableSet(this.selectedKeys);
        }
    }
    
    @Override
    public Set<SelectionKey> keys() {
        if (!this.isOpen() && !Util.atBugLevel("1.4")) {
            throw new ClosedSelectorException();
        }
        return this.publicKeys;
    }
    
    @Override
    public Set<SelectionKey> selectedKeys() {
        if (!this.isOpen() && !Util.atBugLevel("1.4")) {
            throw new ClosedSelectorException();
        }
        return this.publicSelectedKeys;
    }
    
    protected abstract int doSelect(long p0) throws IOException;
    
    private int lockAndDoSelect(long n) throws IOException {
        synchronized (this) {
            if (!this.isOpen()) {
                throw new ClosedSelectorException();
            }
            Set<Object> publicKeys;
            synchronized (publicKeys = this.publicKeys) {
                Set<Object> publicSelectedKeys;
                synchronized (publicSelectedKeys = this.publicSelectedKeys) {
                    return this.doSelect(n);
                }
            }
        }
    }
    
    @Override
    public int select(long n) throws IOException {
        if (n < 0L) {
            throw new IllegalArgumentException("Negative timeout");
        }
        return this.lockAndDoSelect((n == 0L) ? -1L : n);
    }
    
    @Override
    public int select() throws IOException {
        return this.select(0L);
    }
    
    @Override
    public int selectNow() throws IOException {
        return this.lockAndDoSelect(0L);
    }
    
    public void implCloseSelector() throws IOException {
        this.wakeup();
        synchronized (this) {
            synchronized (this.publicKeys) {
                synchronized (this.publicSelectedKeys) {
                    this.implClose();
                }
            }
        }
    }
    
    protected abstract void implClose() throws IOException;
    
    public void putEventOps(SelectionKeyImpl selectionKeyImpl, int n) {
    }
    
    @Override
    protected final SelectionKey register(AbstractSelectableChannel abstractSelectableChannel, int n, Object ob) {
        if (!(abstractSelectableChannel instanceof SelChImpl)) {
            throw new IllegalSelectorException();
        }
        SelectionKeyImpl selectionKeyImpl = new SelectionKeyImpl((SelChImpl)abstractSelectableChannel, this);
        selectionKeyImpl.attach(ob);
        synchronized (this.publicKeys) {
            this.implRegister(selectionKeyImpl);
        }
        selectionKeyImpl.interestOps(n);
        return selectionKeyImpl;
    }
    
    protected abstract void implRegister(SelectionKeyImpl p0);
    
    void processDeregisterQueue() throws IOException {
        Set<SelectionKey> cancelledKeys = this.cancelledKeys();
        synchronized (cancelledKeys) {
            if (!cancelledKeys.isEmpty()) {
                Iterator<Object> iterator = cancelledKeys.iterator();
                while (iterator.hasNext()) {
                    SelectionKeyImpl selectionKeyImpl = (SelectionKeyImpl)iterator.next();
                    try {
                        this.implDereg(selectionKeyImpl);
                    }
                    catch (SocketException cause) {
                        throw new IOException("Error deregistering key", cause);
                    }
                    finally {
                        iterator.remove();
                    }
                }
            }
        }
    }
    
    protected abstract void implDereg(SelectionKeyImpl p0) throws IOException;
    
    @Override
    public abstract Selector wakeup();
}
